'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault(ex) {
  return ex && typeof ex === 'object' && 'default' in ex ? ex['default'] : ex;
}

var loaderUtils = require('loader-utils');
var validate = _interopDefault(require('schema-utils'));

var type = 'object';
var definitions = {};
var properties = {
  export: {
    type: 'string',
    enum: [
      'buffer',
      'instance',
      'module',
      'async',
      'async-instance',
      'async-module'
    ]
  }
};
var additionalProperties = false;
var schema = {
  type: type,
  definitions: definitions,
  properties: properties,
  additionalProperties: additionalProperties
};

/** Wrap binary data as commonjs module so it can be imported by doing require(module)
 * @param buffer raw binary data to be wrapped as es6 module
 * @return chainable object which represent `wrap this data as...`
 * @example return wrap(arrayBuffer).asWebAssembly.Module
 */
function wrap(buffer, module) {
  const data = buffer.toJSON().data.toString();
  let exportString = 'module.exports ='; // if (module === 'cjs')
  if (module === 'esm') exportString = 'export default';
  return {
    asBuffer: `${exportString} Buffer.from([${data}])`,
    asWebAssembly: {
      Module: `${exportString} new WebAssembly.Module(
          Buffer.from([${data}])
        )`,
      Instance: `${exportString} new WebAssembly.Instance(
          new WebAssembly.Module(
            Buffer.from([${data}])
          )
        )`
    },
    promiseWebAssembly: {
      Module: `${exportString} () => WebAssembly.compile(
          Buffer.from([${data}])
        )`,
      Instance: `${exportString} importObject => WebAssembly.instantiate(
          new WebAssembly.Module(Buffer.from([${data}])),
          importObject
        )`,
      Both: `${exportString} importObject => WebAssembly.instantiate(
            Buffer.from([${data}]), importObject
        )`
    }
  };
}

//#region helpers
const is = type => ({
  oneOf: types => types.some(_ => _ === type)
});
function panic(message, cb) {
  if (cb) cb(message);
  else throw new Error(message);
}
//#endregion
/** Transform WebAssembly binary into JS module
 * @param source - WebAssembly Buffer
 * @param options - what type of export you want to generate
 * @return string of the code
 * @example const code = wasm2js(Buffer.from([0x00, 0x61, 0x73, 0x6d, 0x01, 0, 0, 0]), 'instance')
 */
function wasm2js(source, options) {
  if (WebAssembly.validate(source) === false)
    panic('Invalid WebAssembly file', options.errorHandler);
  if (
    is(options.export).oneOf(['instance', 'module']) &&
    source.byteLength > 4000
  )
    panic(
      `The buffer size is larger than 4KB. Consider using async-${
        options.export
      }`,
      options.errorHandler
    );
  switch (options.export) {
    case 'buffer':
      return wrap(source, options.module).asBuffer;
    case 'instance':
      return wrap(source, options.module).asWebAssembly.Instance;
    case 'module':
      return wrap(source, options.module).asWebAssembly.Module;
    case 'async':
      return wrap(source, options.module).promiseWebAssembly.Both;
    case 'async-instance':
      return wrap(source, options.module).promiseWebAssembly.Instance;
    case 'async-module':
      return wrap(source, options.module).promiseWebAssembly.Module;
    default:
      throw new Error(`exporting as "${options.export}" not available`);
  }
}

const defaultOptions = {
  export: 'async'
};
module.exports = function(source) {
  const options = loaderUtils.getOptions(this) || defaultOptions;
  validate(schema, options, 'webassembly-loader');
  // if (options.export === 'buffer') return source; /*🤔*/
  return wasm2js(source, {
    export: options.export,
    module: 'cjs',
    errorHandler: errMsg => this.emitError(errMsg)
  });
};
module.exports.raw = true;

exports.default = wasm2js;
